//;-------------------------------------------------------------------------
//; Copyright (c) 1999 Steven Freyder, David Helland, and Bruce D. Lightner.
//; Use subject to licensing restrictions described in "PICOWEB SERVER
//; SOFTWARE LICENSE AGREEMENT DATED JULY 14, 1999" (See file LICENSE.tt).
//;-------------------------------------------------------------------------

/******************************************************************************

   Programmer for Atmel AVR Microcontrollers (e.g., AT90S84515)

   Author: D.R. Helland

   Revision History:

   990305  BDL  Add more command line options
   990304  BDL  Code cleanup, add load/store for EEPROM, change menu
   990303  BDL  Add command line options, fix EEPROM read/write
   990302  DRH  Debug code, make work on AT90S84515 part
   9902??  DRH  Adopt code from Jeff Bachiochi's BASIC program (FTB101.BAS)

*******************************************************************************

  Parallel Port Programming Cable

  DB-25  8-Pin Header  Signal Name
  -----  ------------  -----------
    2       JP1-1      RESET*
    3       JP1-3      SCK
    4       JP1-7      MOSI
   13       JP1-5      MISO
   15       JP1-2      GND
   24       JP1-4      GND
   23       JP1-6      GND
   22       JP1-8      GND
   12         -         \ Loop-back (cable detect)
   14         -         / Loop-back (cable detect)

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined (__GNUC__)
#include <conio.h>
#include <dos.h>
///#define delay_ms(x) delay(x)
#else
int inportb(int port)
{
    return 0;
};
void outportb(int port, int value)
{
};
void delay_ms(int msecs)
{
};
void getch(void)
{
    getchar();
};

#endif

#ifdef NOT_MSC_COMPAT
#define inp inportb
#define outp outportb
#endif

#define MEM_MAX        0xfff	// 8515 highest Flash WORD address
#define EEPROM_MAX     0x1ff	// 8515 highest EEPROM BYTE address

/* program bytes types */
#define MEM_LO         1
#define MEM_HI         2
#define EEPROM         3

/* ROM file types */
#define ROM_FLASH      100
#define ROM_EEPROM     101

#define ENABLE_1       0xac
#define ENABLE_2       0x53
#define CHIP_ERASE     0x80
#define READ_MEM_HI    0x28
#define READ_MEM_LO    0x20
#define WRITE_MEM_HI   0x48
#define WRITE_MEM_LO   0x40
#define READ_EEPROM    0xa0
#define WRITE_EEPROM   0xc0
#define WRITE_LOCK     0xe0
#define READ_DEV_CODE  0x30

#define SCKHI   0x2
#define RESETHI 0x1
#define MOSIHI  0x4
//#define SCKLO   0
#define RESETLOW  0
#define MISO   0x10

static unsigned char mosi = 0;
static int debug = 0;
static char *program;
static int chip_enabled = 0;
static int reset_active_on_exit = 1;

static unsigned char r1, r2, r3, r4;

static char instring[80];

static int lpt[3] =
{0x03bc, 0x0378, 0x0278};

static int lpt_data = 0;	// 8 DATA BITS

static int lpt_control = 2;	// X X X IRQ4EN *SEL INIT *AUTOFEED *STROBE (POR XXX01011)

static int lpt_status = 1;	// *BUSY *ACK P.OUT SEL *ERR X X X

static int lptN = -1;
static int lpt_port = 0;

static unsigned int mem_lo = 0;
static unsigned int mem_hi = 0;
static unsigned int mem_buf[MEM_MAX + 1];

static long lpt_port_hits_per_msec = 2000L;    /* 1.5-2.0 usec per inp() */

delay_ms(int msecs)
{
    int data;
    register long count;

    count = lpt_port_hits_per_msec * msecs;
    while (count-- > 0L) {
        data = inp(lpt[lpt_port] + lpt_control);
    }
    return data;
}

int find_lpt_port(int port)
{
    int x;
    int first_port = 0;
    int last_port = 2;
    int i;
    int data;

#if !defined (__GNUC__)
    if (port >= 0) {
        first_port = port;
        last_port = port;
    }
    for (x = first_port; x <= last_port; x++) {
        data = inp(lpt[x] + lpt_control);
        outp((lpt[x] + lpt_control), (data | 2));
        i = inp(lpt[x] + lpt_status);
        if (debug)
            printf("LPT%d (0x%x): status=0x%02x\n", x, lpt[x], i);
        if ((i & 0x20) != 0)
            goto skip;
        outp((lpt[x] + lpt_control), (data & (~2)));
        i = inp(lpt[x] + lpt_status);
        if (debug)
            printf("\tstatus=0x%02x\n", i);
        if ((i & 0x20) != 0x20)
            goto skip;
        printf("Found LPT%d (I/O base 0x%x)\n", x, lpt[x]);
        return lpt[x];
      skip:
        i = inp(lpt[x] + lpt_status);
        if (debug)
            printf("\tstatus=0x%02x\n", i);
        outp((lpt[x] + lpt_control), data);
    }
    if (port >= 0) {
        fprintf(stderr, "Can't find the AVR program cable on LPT%d!\n", port);
    } else {
        fprintf(stderr, "Can't find the AVR program cable on any LPT port!\n");
    }
    exit(1);
#endif
    return 0;
}

unsigned char clock_bits(unsigned int out_ch)
{
    unsigned char ret_ch;
    int shift_cnt;
    unsigned char mask = 0x80;
    ret_ch = 0;

    out_ch &= 0xff;
    for (shift_cnt = 0; shift_cnt < 8; shift_cnt++) {
        if ((out_ch & (mask >> shift_cnt)) == (mask >> shift_cnt))
            mosi = MOSIHI;
        else
            mosi = 0;
        outp(lpt_port, mosi);
        outp(lpt_port, mosi + SCKHI);
        if ((inp(lpt_port + lpt_status) & MISO) == MISO) {
            ret_ch = ret_ch + (mask >> shift_cnt);
        }
        outp(lpt_port, mosi);
    }
    return ret_ch;
}

void send_bytes(unsigned int i1, unsigned int i2, unsigned int i3, unsigned int i4)
{
    if (debug)
        printf("[cmd: %02x adr_hi: %02x adr_lo: %02x data: %02x",
               i1 & 0xff, i2 & 0xff, i3 & 0xff, i4 & 0xff);

    r1 = clock_bits(i1);
    r2 = clock_bits(i2);
    r3 = clock_bits(i3);
    r4 = clock_bits(i4);

    if (debug)
        printf(" -> %02x %02x %02x %02x]\n", r1, r2, r3, r4);
}

int prog_mem_byte(int byte_type, int adr_hi, int adr_lo, int val)
{
    int wcmd, rcmd;
    long i;
    char *text;

    val &= 0xff;

    switch(byte_type) {
    case MEM_LO:
        wcmd = WRITE_MEM_LO;
        rcmd = READ_MEM_LO;
        adr_hi &= 0xf;
        break;
    case MEM_HI:
        wcmd = WRITE_MEM_HI;
        rcmd = READ_MEM_HI;
        adr_hi &= 0xf;
        break;
    case EEPROM:
        wcmd = WRITE_EEPROM;
        rcmd = READ_EEPROM;
        adr_hi &= 0x1;
        break;
    default:
        fprintf(stderr, "Error! prog_mem_byte: bad type (0x%x)!\n", byte_type);
        return -1;
    }

#if 0
    /* if writing all ones then check if already 0xff */
    if (val == 0xff) {
        send_bytes(rcmd, adr_hi, adr_lo, 0);
        if (r4 == val)
            return val;
    }
#endif

    /* if writing EEPROM check if already correct */
    if (byte_type == EEPROM) {
        send_bytes(rcmd, adr_hi, adr_lo, 0);
        if (r4 == val)
            return val;
    }

    /* program byte */
    send_bytes(wcmd, adr_hi, adr_lo, val);
#ifdef SLOW_WAY
    delay_ms(4);
#else
    delay_ms(1);   /* this seems to stop "fast write" intermittent errors */
    if (val == 0xff || val == 0x7f) {
        /* chip returns 0x7F until byte write is done...special case */
        delay_ms(4);
        return val;
    } else {
        /* wait for byte to be programmed */
        for (i = 0; i < (debug ? 2 : 10000); ++i) {
            send_bytes(rcmd, adr_hi, adr_lo, 0);
            if (r4 == val)
                return val;
        }
    }
#endif

    /* verify that byte was written correctly */
    send_bytes(rcmd, adr_hi, adr_lo, 0);
    if (r4 != val) {
        switch(byte_type) {
        case MEM_LO: text = "flash memory low"; break;
        case MEM_HI: text = "flash memory high"; break;
        case EEPROM: text = "EEPROM memory"; break;
        }
        fprintf(stderr,
                "Error! Write to %s byte 0x%03x failed! (0x%x!=0x%x)\n",
                text, (adr_hi << 8) | adr_lo, r4, val);
        return -1;
    } else {
        return val;
    }
}

int get_mem_range(
char *text,
char *arg1,
char *arg2,
int base,
unsigned int low,
unsigned int high,
int *begin,
int *end)
{
    int start, finish, n;
    char *p, *fmt, *pbuf, buf[80];

if (debug) printf("get_mem_range(%lx %lx, %lx, %d, %d, %d, %lx, %lx)\n",
(long)*text,
(long)*arg1,
(long)*arg2,
(int)base,
(int)low,
(int)high,
(long)begin,
(long)end);

    *begin = low;
    if (end != NULL) *end = high;
    if (base == 16) {
        fmt = "%s%s (0x%x...0x%x) [0x%x]: ";
    } else {
        fmt = "%s%s (%d...%d) [%d]: ";
    }
    while (1) {
        start = low;		// default to min

        if (arg1 == NULL) {
            printf(fmt,
                (end == NULL) ? "" : "start address", text, low, high, start);
            buf[0] = '\0';
            buf[sizeof(buf) - 1] = '\0';
            if (fgets(buf, sizeof(buf) - 1, stdin) == NULL)
                return 0;
            if ((p = strchr(buf, '\n')) != NULL)
                *p = '\0';
            pbuf = buf;
        } else {
            pbuf = arg1;
        }

        if (*arg1 == '\0')
            break;		// RETURN gets default

        start = (int)strtol(pbuf, &p, base);
        if ((p == NULL) || (*p != '\0')) {
            printf("Non-numeric input!\n");
            return 0;
        }
        if ((start < low) || (start > high)) {
            printf("Illegal %s %d! (must be %d...%d)\n",
                   (end == NULL) ? "value" : "start address",
                   start, low, high);
            if (arg1 != NULL) return 0;
            continue;
        }
        break;
    }
    *begin = start;
    if (end == NULL) return 1;

    while (1) {
        finish = high;		// default to min

        if (arg2 == NULL) {
            printf(fmt, "ending address", text, start, high, finish);
            buf[0] = '\0';
            buf[sizeof(buf) - 1] = '\0';
            if (fgets(buf, sizeof(buf) - 1, stdin) == NULL)
                return 0;
            if ((p = strchr(buf, '\n')) != NULL)
                *p = '\0';
            if (*buf == '\0')
                break;		// RETURN gets default
            pbuf = buf;
        } else {
            pbuf = arg2;
        }

        finish = (int)strtol(pbuf, &p, 0);
        if ((p == NULL) || (*p != '\0')) {
            printf("Non-numeric input!\n");
            return 0;
        }
        if ((finish < start) || (finish > high)) {
            printf("Illegal finish address %d! (must be %d...%d)\n",
                   finish, start, high);
            if (arg2 != NULL) return 0;
            continue;
        }
        break;
    }
    *end = finish;
    return 1;
}

int burn_memory(int rom_type)
{
    int i, addr;
    int lo = 0, hi;
    unsigned int val;
    unsigned int adr_hi, adr_lo, val_hi, val_lo;
    char *rom_name;

    switch(rom_type) {
    case ROM_FLASH: hi = MEM_MAX; rom_name = "Flash"; break;
    case ROM_EEPROM: hi = EEPROM_MAX; rom_name = "EEPROM"; break;
    default:
        fprintf(stderr, "Error! burn_memory: bad type (0x%x)!\n", rom_type);
        return 0;
    }

    printf("writing %s memory (%d-%d)\n", rom_name, mem_lo, mem_hi);
    fflush(stdout);
    if (mem_lo < lo || mem_hi > hi) {
        fprintf(stderr, "Error! Flash memory buffer not in range %d...%d!!\n",
            lo, hi);
        return 0;
    }

    i = 0;
    for (addr = mem_lo; addr <= mem_hi; ++addr) {
        val = mem_buf[addr];
        adr_hi = (addr >> 8) & 0xff;
        adr_lo = addr & 0xff;
        val_hi = (val >> 8) & 0xff;
        val_lo = val & 0xff;

        switch(rom_type) {
        case ROM_FLASH:
            if (prog_mem_byte(MEM_LO, adr_hi, adr_lo, val_lo) < 0) {
                fflush(stdout);
                return 0;
            }
            if (prog_mem_byte(MEM_HI, adr_hi, adr_lo, val_hi) < 0) {
                fflush(stdout);
                return 0;
            }
            break;
        case ROM_EEPROM:
            if (prog_mem_byte(EEPROM, adr_hi, adr_lo, val_lo) < 0) {
                fflush(stdout);
                return 0;
            }
            break;
        }
        if ((i++ % 100) == 0) {
            printf(".");
            fflush(stdout);
        }
    }
    printf("\n");
    return 1;
}

int read_rom_file(int rom_type, char *filename)
{
    char hex_str[80];
    char hex_val[10];
    int linenum = 0;
    int i, addr, val, len;
    int lo = 0, hi;
    unsigned int adr_hi, adr_lo, val_hi, val_lo;
    char *p;
    FILE *in;

    switch(rom_type) {
    case ROM_FLASH: hi = MEM_MAX; len = 11; break;
    case ROM_EEPROM: hi = EEPROM_MAX; len = 7; break;
    default:
        fprintf(stderr, "Error! read_rom_file: bad type (0x%x)!\n", rom_type);
        return 0;
    }

    if ((in = fopen(filename, "rt")) == NULL) {
        fprintf(stderr, "Cannot open input file %s!\n", filename);
        perror(filename);
        return 0;
    }
    mem_lo = mem_hi = 0xffff;
    while (!feof(in)) {
        fgets(hex_str, 20, in);
        if ((p = strchr(hex_str, '\n')) != NULL)
            *p = '\0';
        if ((p = strchr(hex_str, '\r')) != NULL)
            *p = '\0';
        ++linenum;
        if (strlen(hex_str) != len) {
            printf("Bad data at %s line %d\n", filename, linenum);
            printf("   Bad input : %s\n", hex_str);
            printf("   Bad len   : %u (expected %d)\n", strlen(hex_str), len);
            return 0;
        }
        if (debug)
            printf("%s \n", hex_str);
        switch(rom_type) {
        case ROM_FLASH:
            i = sscanf(hex_str, "%4x%2x:%x", &adr_hi, &adr_lo, &val);
            break;
        case ROM_EEPROM:
            i = sscanf(hex_str, "%2x%2x:%x", &adr_hi, &adr_lo, &val);
            break;
        }
        if (i != 3) {
            printf("Bad format at %s line %d (%d)\n", filename, linenum, i);
            printf("   Bad input : %s\n", hex_str);
            return 0;
        }
        addr = (adr_hi << 8) | adr_lo;
        val_lo = val & 0xff;
        val_hi = (val >> 8) & 0xff;
        if (addr < 0 || addr > hi) {
            printf("Bad address at %s line %d\n", filename, linenum);
            printf("   Bad input : %s\n", hex_str);
            printf("   Bad addr  : 0x%x (expected %d...%d)\n", addr, lo, hi);
            return 0;
        }
        mem_buf[addr] = (val_hi << 8) | val_lo;
        if (mem_hi == 0xffff)
            mem_lo = mem_hi = addr;
        if (addr < mem_lo)
            mem_lo = addr;
        if (addr > mem_hi)
            mem_hi = addr;
    }
    return 1;
}

int chip_enable(void)
{
    int retry;

    if (!lpt_port)
        lpt_port = find_lpt_port(lptN);

    if (chip_enabled)
        return 1;

    if (debug) printf("Sending the programming enable command\n");
    r3 = 0;
    retry = 0;

    outp(lpt_port, RESETLOW);
    delay_ms(20);
    outp(lpt_port, RESETHI);
    delay_ms(20);
    outp(lpt_port, RESETLOW);
    delay_ms(20);

    while ((r3 != ENABLE_2) && (retry++ < 10)) {

        send_bytes(ENABLE_1, ENABLE_2, 0, 0);
        delay_ms(4);
        if (debug)
            printf("r3=%x \n", r3);
        if (r3 == ENABLE_2)
            break;

        outp(lpt_port, SCKHI + RESETLOW);
        delay_ms(4);
        outp(lpt_port, RESETLOW);
        delay_ms(4);
    }
#if !defined (__GNUC__)
    if (r3 != ENABLE_2) {
        printf("Chip can not be enabled...\n");
        outp(lpt_port, RESETLOW);
        exit(1);
    }
#endif
    if (debug) printf("Chip enabled \n");
    chip_enabled = 1;
    return 1;
}

void chip_erase(void)
{
    printf("Sending the chip ERASE command\n");
    send_bytes(ENABLE_1, CHIP_ERASE, 0, 0);
    delay_ms(20);
    outp(lpt_port, RESETHI);
    delay_ms(20);
    chip_enabled = 0;
    chip_enable();
    return;
}

void read_flash(int start, int finish)
{
    int n;

    mem_lo = start;
    mem_hi = finish;
    for (n = start; n <= finish; n++) {
        if (debug)
            printf("%4.4x : ", n);
        send_bytes(READ_MEM_HI, (n >> 8), n, 0);
        ////delay_ms(4);
        if (debug)
            printf("%2.2x ", r4);
        mem_buf[n] = r4;
        mem_buf[n] = mem_buf[n] << 8;

        send_bytes(READ_MEM_LO, (n >> 8), n, 0);
        ////delay_ms(4);
        if (debug)
            printf("%2.2x\n", r4);
        mem_buf[n] = mem_buf[n] | r4;
    }
    return;
}

void read_EEPROM(int start, int finish)
{
    int n;

    mem_lo = start;
    mem_hi = finish;
    for (n = start; n <= finish; n++) {
        if (debug)
            printf("%4.4x : ", n);
        send_bytes(READ_EEPROM, (n >> 8), n, 0);
        ////delay_ms(4);
        if (debug)
            printf("%2.2x ", r4);
        mem_buf[n] = r4 & 0xff;
    }
    return;
}

void list_flash(int start, int finish)
{
    int i, n;

#if !defined (__GNUC__)
    for (n = start; n <= finish; n++) {
        if (debug)
            printf("%4.4x : ", n);
        send_bytes(READ_MEM_HI, (n >> 8), n, 0);
        ////delay_ms(4);
        if (debug)
            printf("%2.2x ", r4);
        mem_buf[n] = r4;
        mem_buf[n] = mem_buf[n] << 8;

        send_bytes(READ_MEM_LO, (n >> 8), n, 0);
        ////delay_ms(4);
        if (debug)
            printf("%2.2x\n", r4);
        mem_buf[n] = mem_buf[n] | r4;
    }
#endif

    i = 0;
    for (n = start; n <= finish; n++) {
        if ((i++ % 8) == 0)
            printf("\n%4.4x :", n);
        printf(" %04x", mem_buf[n]);
    }
    printf("\n");
    return;
}

void list_EEPROM(int start, int finish)
{
    int i, n;

    i = 0;
    for (n = start; n <= finish; n++) {
        send_bytes(READ_EEPROM, (n >> 8), n, 0);
        ///delay_ms(4);
        if ((i++ % 16) == 0)
            printf("\n%3.3x :", n);
        printf(" %02x", r4);
    }
    printf("\n");
    return;
}

int write_EEPROM_byte(void)
{
    int start, data_val;

    if (!get_mem_range("EEPROM write address", NULL, NULL, 0, 0, EEPROM_MAX, &start, NULL))
        return 0;

    if (!get_mem_range("8-bit byte to write", NULL, NULL, 16, 0, 0xff, &data_val, NULL))
        return 0;

    mem_buf[start] = data_val;

    if (prog_mem_byte(EEPROM, (start >> 8), start, data_val) < 0)
        return 0;

    return 1;
}

int write_flash_word(void)
{
    int start, data_val;

    if (!get_mem_range("Flash memory write address", NULL, NULL, 0, 0, MEM_MAX, &start, NULL))
        return 0;

    if (!get_mem_range("16-bit word to write", NULL, NULL, 16, 0, 0xffff, &data_val, NULL))
        return 0;

    mem_buf[start] = data_val;

    if (prog_mem_byte(MEM_LO, (start >> 8), start, data_val) < 0)
        return 0;

    if (prog_mem_byte(MEM_HI, (start >> 8), start, (data_val >> 8)) < 0)
        return 0;

    return 1;
}

void write_lock(void)
{
    int again;

  try_again:
    printf("\nChoose a LOCK bit command\n");
    printf(" 1 - Disable writing of program and EEPROM memory\n");
    printf(" 2 - Disable writing and verification of program and EEPROM memory\n");
    printf(" 3 - Don't do anything\n");
    printf(": ");
    gets(instring);
    again = 0;
    switch (instring[0]) {
    case '1':
        break;
    case '2':
        break;
    case '3':
        return;
    default:
        again = 1;
    }
    if (again)
        goto try_again;

    if (instring[0] == '1') {
        printf("Disable writing of program and EEPROM memory...\n");
        printf("Are you sure (Y/N)? ");
        gets(instring);
        if ((instring[0] == 'y') || (instring[0] == 'Y')) {
            printf("Sending command\n");
            send_bytes(ENABLE_1, 0xfd, 0, 0);
            delay_ms(4);
        }
    }
    if (instring[0] == '2') {
        printf("Disable writing and verification of program and EEPROM memory...\n");
        printf("Are you sure (Y/N)? ");
        gets(instring);
        if ((instring[0] == 'y') || (instring[0] == 'Y')) {
            printf("Sending command\n");
            send_bytes(ENABLE_1, 0xf9, 0, 0);
            delay_ms(4);
        }
    }
    return;
}

void read_dev_codes(void)
{
    unsigned char n;

    printf("Device codes:");
    for (n = 0; n < 3; n++) {
        send_bytes(READ_DEV_CODE, 0, n, 0);
        delay_ms(4);
        printf(" %02x", r4);
    }
    printf("\n");
    return;
}

int write_data_file(int rom_type, char *fname, unsigned int *buf, int start, int finish)
{
    char hex_str[80];
    FILE *out;
    unsigned int addr;

    switch(rom_type) {
    case ROM_FLASH: break;
    case ROM_EEPROM: break;
    default:
        fprintf(stderr, "Error! write_data_file: bad type (0x%x)!\n", rom_type);
        return 0;
    }

    if ((out = fopen(fname, "wt")) == NULL) {
        fprintf(stderr, "Cannot open output file %s!\n", fname);
        perror(fname);
        return 0;
    }

    for (addr = start; addr <= finish; addr++) {
        switch(rom_type) {
        case ROM_FLASH:
            sprintf(hex_str, "%6.6x:%4.4x\n", addr, buf[addr]);
            break;
        case ROM_EEPROM:
            sprintf(hex_str, "%4.4x:%2.2x\n", addr, buf[addr]);
            break;
        }
        fputs(hex_str, out);
    }
    fclose(out);
    switch(rom_type) {
    case ROM_FLASH: printf("Program memory "); break;
    case ROM_EEPROM: printf("EEPROM "); break;
    }
    printf("file %s written (%d-%d)\n", fname, mem_lo, mem_hi);

    return 1;
}

void usage(void)
{
    fprintf(stderr, "usage: %s [options]\n"
    "options:\n"
    "   -d         - set debug mode\n"
    "   -lptx      - look for program cable on parallel port LPTx\n"
    "   -xNNN      - program delay factor (default %ld)\n"
    "   -ce        - Chip Erase\n"
    "   -en        - Enable chip (deassert RESET)\n"
    "   -lp file   - Load Program memory from file\n"
    "   -le file   - Load EEPROM memory from file\n"
    "   -sp file   - Save Program memory to file\n"
    "   -se file   - Save EEPROM memory to file\n"
    "   -dp n1 n2  - Display Program memory in range n1...n2\n"
    "   -de n1 n2  - Display EEPROM memory in range n1...n2\n"
    "   -dd        - Display Device codes\n"
    "   -ap addr w - Alter Program memory at addr with  16-bit word w\n"
    "   -ae addr b - Alter EEPROM memory at addr with 8-bit byte b\n"
    "   -wp        - Write Protect Program/EEPROM memory\n"
    "   -rp        - Read/write Protect Program/EEPROM memory\n"
        , program, lpt_port_hits_per_msec);
    exit(1);
}

void cleanup()
{
    int i;

    if (reset_active_on_exit) {
        outp(lpt_port, RESETLOW);
    } else {
        /* delay about 1 msec with RESET active */
        for (i = 0; i < 1000; ++i) {  
            outp(lpt_port, RESETLOW);
	}
        outp(lpt_port, RESETHI);
    }
}

void main(int argc, char **argv)
{
    int n;
    int start, finish, data_val;
    char *p, *opt;
    char filename[256];

    program = *argv;
    --argc;
    ++argv;
    if ((p = strrchr(program, '/')) != NULL)
        program = p + 1;
    if ((p = strrchr(program, '\\')) != NULL)
        program = p + 1;

    for (n = 0; n < MEM_MAX; n++)
        mem_buf[n] = 0xffff;

    atexit(cleanup);

    while (argc && **argv == '-') {
        opt = *argv;
        --argc;
        ++argv;
        if (!strcmp(opt, "-d")) {
            debug = 1;

        } else if (!strncmp(opt, "-x", 2)) {
            lpt_port_hits_per_msec = (long)atoi(opt + 2);
            if (lpt_port_hits_per_msec <= 0) usage();

        } else if (!strncmp(opt, "-lpt", 4)) {
            lptN = atoi(opt + 4);
            if (lptN < 0 || lptN > 2) {
                fprintf(stderr, "%s: no such LPT port (%s)!  Use -lpt0, -lpt1 or -lpt2.\n", program, *argv);
                exit(1);
            }
            lpt_port = find_lpt_port(lptN);

        } else if (!strcmp(opt, "-ce")) {
            chip_enable();
            chip_erase();

        } else if (!strcmp(opt, "-en")) {
            chip_enable();
            reset_active_on_exit = 0;

        } else if (!strcmp(opt, "-lp")) {
            if (argc < 1) usage();
            chip_enable();
            p = *argv; --argc; ++argv;
            if (read_rom_file(ROM_FLASH, p))
                burn_memory(ROM_FLASH);

        } else if (!strcmp(opt, "-le")) {
            if (argc < 1) usage();
            p = *argv; --argc; ++argv;
            if (read_rom_file(ROM_EEPROM, p))
                chip_enable();
                burn_memory(ROM_EEPROM);

        } else if (!strcmp(opt, "-sp")) {
            if (argc < 1) usage();
            p = *argv; --argc; ++argv;
            chip_enable();
            read_flash(0, MEM_MAX);
            for (mem_hi = MEM_MAX; mem_hi != 0; --mem_hi) {
               if (mem_buf[mem_hi] != 0xffff)
                   break;
            }
            write_data_file(ROM_FLASH, p, mem_buf, mem_lo, mem_hi);

        } else if (!strcmp(opt, "-se")) {
            if (argc < 1) usage();
            p = *argv; --argc; ++argv;
            chip_enable();
            read_EEPROM(0, EEPROM_MAX);
#ifdef COMPRESS_EEPROM
            for (mem_hi = EEPROM_MAX; mem_hi != 0; --mem_hi) {
               if (mem_buf[mem_hi] != 0xff)
                   break;
            }
#endif
            write_data_file(ROM_EEPROM, p, mem_buf, mem_lo, mem_hi);

        } else if (!strcmp(opt, "-dp")) {
            if (argc < 2) usage();
            if (!get_mem_range("", *argv, *(argv + 1), 0, 0, MEM_MAX, &start, &finish))
                exit(1);
            chip_enable();
            list_flash(start, finish);
            argc -= 2; argv += 2;

        } else if (!strcmp(opt, "-de")) {
            if (argc < 2) usage();
            if (!get_mem_range("", *argv, *(argv + 1), 0, 0, EEPROM_MAX, &start, &finish))
                exit(1);
            chip_enable();
            list_EEPROM(start, finish);
            argc -= 2; argv += 2;

        } else if (!strcmp(opt, "-dd")) {
            chip_enable();
            read_dev_codes();

        } else if (!strcmp(opt, "-ap")) {
            if (argc < 2) usage();
            if (!get_mem_range("", *argv, NULL, 0, 0, MEM_MAX, &start, NULL))
                exit(1);
            --argc; ++argv;
            if (!get_mem_range("", *argv, NULL, 16, 0, 0xffff, &data_val, NULL))
                exit(1);
            --argc; ++argv;
            chip_enable();
            if (prog_mem_byte(MEM_LO, (start >> 8), start, data_val) < 0)
                exit(1);
            if (prog_mem_byte(MEM_HI, (start >> 8), start, (data_val >> 8)) < 0)
                exit(1);

        } else if (!strcmp(opt, "-ae")) {
            if (argc < 2) usage();
            if (!get_mem_range("", *argv, NULL, 0, 0, EEPROM_MAX, &start, NULL))
                exit(1);
            --argc; ++argv;
            if (!get_mem_range("", *argv, NULL, 16, 0, 0xff, &data_val, NULL))
                exit(1);
            --argc; ++argv;
            chip_enable();
            if (prog_mem_byte(EEPROM, (start >> 8), start, data_val) < 0)
                exit(1);

        } else if (!strcmp(opt, "-wp")) {
            chip_enable();
            send_bytes(ENABLE_1, 0xfd, 0, 0);  // lock bit 2 = 0
            delay_ms(4);
            printf("Sent write protect bit\n");

        } else if (!strcmp(opt, "-rp")) {
            chip_enable();
            send_bytes(ENABLE_1, 0xf9, 0, 0);  // lock bits 1,2 = 0
            delay_ms(4);
            printf("Sent write protect and read protect bits\n");

        } else {
            fprintf(stderr, "%s: unknown option (%s)!\n", program, opt);
            usage();
        }
    }

    if (argc)
        usage();

    if (chip_enabled)
        exit(0);

    printf("Placing chip into RESET\n");
    outp(lpt_port, RESETLOW);
    delay_ms(100);

    chip_enable();

  menu:

    printf("\n");
    printf(" CE - Chip Erase\n");
    printf(" WL - Write Lock bits\n");
    printf(" LP - Load Program memory from file\n");
    printf(" LE - Load EEPROM memory from file\n");
    printf(" SP - Save Program memory to file\n");
    printf(" SE - Save EEPROM memory to file\n");
    printf(" DP - Display Program memory\n");
    printf(" DE - Display EEPROM memory\n");
    printf(" DD - Display Device codes\n");
    printf(" AP - Alter Program memory\n");
    printf(" AE - Alter EEPROM memory\n");
    printf(" Q  - Quit program\n");
    printf("\n");

  again:
    fflush(stderr);
    printf("Enter command: ");
    fflush(stdout);
    instring[0] = '\0';
    gets(instring);
    for (n = 0; n < strlen(instring); ++n) {
        int ch = instring[n];
        if (ch >= 'a' && ch <= 'z')
            instring[n] = 'A' + (ch - 'a');
    }

#   define CHECK(n) (!strcmp(instring,n))
    if        (CHECK("")) {
        goto menu;

    } else if (CHECK("CE")) {
        printf("\nChip ERASE...\nAre you sure (Y/N)? ");
        gets(instring);
        if ((instring[0] == 'y') || (instring[0] == 'Y')) {
            chip_erase();
        }

    } else if (CHECK("WL")) {
        write_lock();

    } else if (CHECK("LP")) {
        printf("Enter input program ROM file name (.ROM): ");
        gets(filename);
        if (read_rom_file(ROM_FLASH, filename))
            burn_memory(ROM_FLASH);

    } else if (CHECK("LE")) {
        printf("Enter input EEPROM file name (.EEP): ");
        gets(filename);
        if (read_rom_file(ROM_EEPROM, filename))
            burn_memory(ROM_EEPROM);

    } else if (CHECK("SP")) {
        printf("Enter output ROM file name (.ROM): ");
        gets(filename);
        if (!get_mem_range("Program memory save", NULL, NULL, 0, 0, MEM_MAX, &start, &finish))
            return;
        read_flash(start, finish);
        write_data_file(ROM_FLASH, filename, mem_buf, mem_lo, mem_hi);

    } else if (CHECK("SE")) {
        printf("Enter output ROM file name (.EEP): ");
        gets(filename);
        if (!get_mem_range("EEPROM memory save", NULL, NULL, 0, 0, EEPROM_MAX, &start, &finish))
            return;
        read_EEPROM(start, finish);
        write_data_file(ROM_EEPROM, filename, mem_buf, mem_lo, mem_hi);

    } else if (CHECK("DP")) {
        if (!get_mem_range("Flash memory list", NULL, NULL, 0, 0, MEM_MAX, &start, &finish))
            return;
        list_flash(start, finish);

    } else if (CHECK("DE")) {
        if (!get_mem_range("EEPROM memory list", NULL, NULL, 0, 0, EEPROM_MAX, &start, &finish))
            return;
        list_EEPROM(start, finish);

    } else if (CHECK("DD")) {
        read_dev_codes();

    } else if (CHECK("AP")) {
        write_flash_word();

    } else if (CHECK("AE")) {
        write_EEPROM_byte();

    } else if (CHECK("DEBUG")) {
        debug = (debug + 1) & 1;
        printf( "debug %s\n", debug ? "on" : "off");

    } else if (CHECK("Q")) {
        outp(lpt_port, RESETLOW);
        printf("Reset is active\n");
        exit(0);

    } else {
        printf("Unknown command (%s)!%c\n", instring, 7);
    }
    goto again;
}

